# manifold.py - Data handling routines developed by Optibrium for integration with the Manifold API
# Use of this code is subject to the terms of the StarDrop License Agreement and use, copying or redistribution of this file for use by other persons or organisations is prohibited
# Copyright(C) Optibrium Ltd 2023
# stardrop-support@optibrium.com
#

"""Module to control access to Manifold from StarDrop"""
import logging
import webbrowser
from logging import Logger

import manifoldconfig
import app
import requests
import PythonQt
from typing import Optional, Union


def _setup_manifold_logger():
    """Set up a logger object for this script.
    """
    log_console_format = "[%(asctime)s] - [%(levelname)s]: %(message)s"
    logger_object: Logger = logging.getLogger(__name__)
    logger_object.setLevel(manifoldconfig.LOGGING_LEVEL)
    if not logger_object.handlers:
        console_handler = logging.FileHandler(manifoldconfig.LOGGER_FILE)
        console_handler.setLevel(manifoldconfig.LOGGING_LEVEL)
        formatter = logging.Formatter(fmt=log_console_format, datefmt='%d-%b-%y %H:%M:%S')
        console_handler.setFormatter(formatter)
        logger_object.addHandler(console_handler)
    logger_object.info(manifoldconfig.version)
    return logger_object


LOGGER = _setup_manifold_logger()


def change_token(appui) -> None:
    """Respond to a user request to change the API token being used.
    @param: appui: The StarDrop application interface.
    """
    dialog = PythonQt.ChangeTokenDialog()
    if dialog.run():  # The user selects "OK"
        new_token = dialog.newToken()
        if new_token == "":
            LOGGER.info("Token is empty, so has not been set.")
            appui.showError("No API token was entered.\nThe API token has not been updated.")
            return
        LOGGER.debug("Token now set")
        manifoldconfig.set_token(new_token)
    else:
        if dialog.isTokenRequested():  # The user selects "Request Token"
            LOGGER.info("User requested API token from PostEra.")
            google_form_url = "https://docs.google.com/forms/d/e/1FAIpQLScJBbvmb5VkFi9KF8hNzDwWIwonvZYOGjebzbAJlxcSBapByw/viewform"
            webbrowser.open(google_form_url)
        else:  # The user selects "Cancel"
            LOGGER.debug("User cancelled the dialog to change the API token.")


def get_fast_score_and_synthetic_accessibility_alert(appui) -> None:
    """Get the fast score data for the selected row in the data set, and add the data as a new column.
    @param: appui: The StarDrop application interface.
    """
    LOGGER.info("Retrieving fast score data from Manifold, for one row in the data set.")
    if not Helper.connect_to_manifold(appui):
        return None
    num_of_selected_rows = len(appui.getCurrentMolecules())
    if num_of_selected_rows > 1:
        Helper.warning_message(message="Too many rows have been selected for this action. \n"
                                       "Ensure that only one row is selected, and try again.")
        return None

    smiles_string, selected_row = Helper.extract_selected_smiles_string(appui)

    if smiles_string:
        LOGGER.debug(f"The SMILES string that has been extracted is: {smiles_string}")
        json_data = {'smiles': smiles_string, 'getAlertSvg': True}

        response_json = Helper.perform_post_request(manifoldconfig.FAST_SCORE_POST_URL, json_data)
        LOGGER.debug("Fast score post request has been sent.")
        fast_score_value, alert, svg = FastScores.process_data(response_json)
        if fast_score_value and svg:
            manifold_display_data = PythonQt.ManifoldDisplayData()
            manifold_display_data.loadSVGData(str(svg))
            manifold_display_data.run()
            appui.startMacro("")
            FastScores.add_to_data_set(appui, fast_score_value, alert, selected_row, smiles_string)
            appui.stopMacro("")
        else:
            Helper.warning_message(title="Processing Error",
                                   message="Unable to process the data retrieved from Manifold.")
    else:
        LOGGER.debug("No SMILES string has been extracted.")
        Helper.warning_message(message="No molecule was found. Select a row which includes a structure and try again.")


def get_fast_scores(appui) -> None:
    """Get the fast score data for all selected rows in the current data set, and add the data as a new column.
    @param: appui: The StarDrop application interface.
    """
    LOGGER.info("Retrieving fast score data from Manifold")
    if not Helper.connect_to_manifold(appui):
        return None

    progress_bar = PythonQt.SimpleProgress()
    progress_bar.setWindowTitle("Retrieving Fast Score Data")
    progress_bar.setMessage("Extracting data from StarDrop...")
    progress_bar.setProgress(16)
    progress_bar.show()

    smiles_list, index_list = Helper.extract_selected_smiles_strings(appui)
    if smiles_list:
        split_smiles_lists = Helper.distribute_list_into_chunks(smiles_list, 100)

        progress_bar.setProgress(32)
        response_data = []

        for lst in split_smiles_lists:
            json_data = {'smilesList': lst}
            response = Helper.perform_post_request(
                manifoldconfig.FAST_SCORE_BATCH_POST_URL,
                json_data
            )
            response_data.append(response)
            LOGGER.debug("A post request for fast score data has been sent.")

        if response_data:
            progress_bar.setProgress(48)
            LOGGER.debug("A response from the fast score post request has been received.")
            processed_data = FastScores.process_batch_data(response_data)
            if not processed_data:
                progress_bar.hide()
                return
            progress_bar.setProgress(64)
            fast_score_data = FastScores.map_smiles_and_score(smiles_list,
                                                              processed_data)
            if fast_score_data:
                progress_bar.setProgress(80)
                appui.startMacro("")
                FastScores.add_batch_to_data_set(appui, fast_score_data, index_list)
                appui.stopMacro("")
            else:
                Helper.warning_message(title="Processing Error",
                                       message="Unable to process the data retrieved from Manifold.")
    else:
        LOGGER.info("No smiles were extracted.")
        Helper.warning_message(message="No molecules could be extracted from the data set.")

    progress_bar.hide()


def get_retrosynthesis_data(appui) -> None:
    """Get the retrosynthesis url link for the selected row, and add the data to a column in the data set.
    @param: appui: The StarDrop application interface.
    """
    LOGGER.info("Retrieving retrosynthesis data from Manifold")
    if not Helper.connect_to_manifold(appui):
        return None

    progress_bar = PythonQt.SimpleProgress()
    progress_bar.setWindowTitle("Retrieving Retrosynthesis Data")
    progress_bar.setMessage("Extracting data from StarDrop...")
    progress_bar.setProgress(10)
    progress_bar.show()

    smiles_list, index_list = Helper.extract_selected_smiles_strings(appui)
    if smiles_list:
        if len(smiles_list) > 10:
            progress_bar.hide()
            Helper.warning_message(title="Exceeded Recommended Compound limit",
                                   message="Please select a maximum of "
                                           "10 rows, and try again.")
            return None

        progress_bar.setProgress(15)

        parameters = Retrosynthesis.retrieve_retrosynthesis_parameters()
        if parameters:
            json_data = parameters
        else:
            LOGGER.debug("User cancelled the action when selecting parameters.")
            return None

        progress_bar.setProgress(20)
        progress_bar.setMessage("Communicating with Manifold server...")

        json_data['smilesList'] = smiles_list
        response_json = Helper.perform_post_request(manifoldconfig.RETROSYNTHESIS_POST_URL, json_data)
        LOGGER.debug("Post request has been sent")
        LOGGER.debug(f"Response is {response_json}")

        progress_bar.setProgress(50)
        progress_bar.setMessage("Processing Manifold data...")

        manifold_scores, manifold_min_num_steps, manifold_links = Retrosynthesis.process_data(response_json, progress_bar)
        progress_bar.setMessage("Adding data to current data set...")
        if manifold_scores and manifold_links:
            LOGGER.debug("Manifold data has been extracted. Adding to data set.")
            appui.startMacro("")
            Retrosynthesis.add_to_data_set(appui, manifold_scores, manifold_min_num_steps, manifold_links, index_list, progress_bar)
            appui.stopMacro("")

    return


class FastScores:
    @staticmethod
    def process_batch_data(response_data: list[dict]) -> list:
        """Process the json data retrieved from Manifold, into a format to input into StarDrop.
        @param: response_data: The json data, retrieved from Manifold.
        @return: The processed data.
        """
        LOGGER.debug(f"Processing fast score data: {response_data}")
        error_count = 0
        fast_score_data = []
        for response in response_data:
            if not Helper.is_api_call_valid(response):
                break
            for json_data in response['results']:
                if 'error' in json_data:
                    error_count += 1
                    fast_score_data.append(None)
                    LOGGER.debug(f"Error detected in response json: {json_data['error']}.")
                else:
                    fast_score = json_data['SAData']['fastSAScore']
                    alert = json_data['SAData']['SAAlertLevel']
                    fast_score_data.append((fast_score, alert))
        if error_count > 0:
            LOGGER.warning(f"{error_count=}")
            if error_count == 1:
                warning_message = f"{error_count} entry has an invalid SMILES string."
            else:
                warning_message = f"{error_count} entries have invalid SMILES strings."
            Helper.warning_message(message=warning_message)
        else:
            if fast_score_data:
                LOGGER.debug("No errors were detected in the response json.")
        return fast_score_data

    @staticmethod
    def process_data(response_data: dict) -> tuple[str, str]:
        """Process the json data retrieved from Manifold, into a format to input into StarDrop.
        @param: response_data: The json data, retrieved from Manifold.
        @return: The processed data.
        """
        LOGGER.debug("Processing the data returned from Manifold.")
        if not Helper.is_api_call_valid(response_data):
            return '', ''
        fast_score = ""
        alert = ""
        svg = None
        if 'fastSAScore' in response_data:
            fast_score = response_data['fastSAScore']
        if 'SAAlertLevel' in response_data:
            alert = response_data['SAAlertLevel']
        if 'SAAlertSvg' in response_data:
            svg = response_data['SAAlertSvg']
        if not fast_score and not svg:
            LOGGER.debug("No data was successfully processed.")
        if not fast_score:
            LOGGER.debug("No fast score data was successfully processed.")
        if not svg:
            LOGGER.debug("No SVG data was successfully processed.")
        return fast_score, alert, svg

    @staticmethod
    def map_smiles_and_score(smiles_list: list, fast_scores: list) -> list:
        """Maps the SMILES and score data together, to align the data.
        @param: smiles_list: The list of SMILES strings.
        @param: fast_score: The list of fast score values.
        @return: The mapped data.
        """
        LOGGER.debug("Map SMILES to fast score")
        if len(smiles_list) == len(fast_scores):
            mapped_data = [(smiles, fast_score[0], fast_score[1])
                           for smiles, fast_score
                           in zip(smiles_list, fast_scores)]
        else:
            LOGGER.debug(f"Smiles list and number of fast scores are not equal. There are {len(smiles_list)} smiles "
                         f"strings, and {len(fast_scores)} fast scores.")
            return []
        LOGGER.debug(f"Mapped fast score data: {mapped_data}")
        return mapped_data

    @staticmethod
    def add_to_data_set(appui, data: str, alert: str, selected_row: int, smiles: str) -> None:
        """Add the fast score data, from the single molecule call, to the data set.
        @param: appui: The StarDrop application interface.
        @param: data: The data to add to the data set.
        @param: alert: The alert string from the API.
        @param: selected_row: The index of the row currently selected in the data set.
        @param: smiles: The SMILES string for the selected row
        """
        LOGGER.debug("Add to data set")
        data_set_bytes = appui.importDS("")
        data_set = app.Dataset()
        data_set.parse(data_set_bytes)

        header_name = "Fast SA Score"
        alert_header_name = "Fast SA Alert"
        column_index = data_set.columnIndex(header_name)
        alert_column_index = data_set.columnIndex(alert_header_name)
        links_header_name = "Manifold Link"
        links_header_index = data_set.columnIndex(links_header_name)
        LOGGER.debug(f"The column index for {header_name} is {column_index}.")
        LOGGER.debug(f"The column index for {alert_header_name} is {alert_column_index}.")

        if not StarDropHandler.column_exists(data_set, header_name):
            LOGGER.debug(f"Creating column {header_name}.")
            data_set = FastScores.construct_column(data_set, header_name)
            LOGGER.debug(f"The value for row {selected_row} is {data}.")
            entry = app.Dataset.cont(data, 0.0)
            data_set.set(column_index, selected_row, entry)
            column_bytes = data_set.packColumn(column_index)
            LOGGER.debug("Adding the column to the data set.")
            appui.addColumnToDataSet(header_name, column_index, column_bytes)
        else:
            LOGGER.debug(f"Column {header_name} already exists.")
            entry = {"value": data}
            appui.updateDataEntry("", selected_row, column_index, entry)
            LOGGER.debug(f"Updated entry for row {selected_row}")
        if not StarDropHandler.column_exists(data_set, alert_header_name):
            LOGGER.debug(f"Creating column {alert_header_name}.")
            data_set, categories_dict = FastScores.construct_category_column(data_set, alert_header_name)
            entry = app.Dataset.createCategoryEntry(alert, categories_dict)
            data_set.set(alert_column_index, selected_row, entry)
            LOGGER.debug(f"The value for row {selected_row} is {data}.")
            column_bytes = data_set.packColumn(alert_column_index)
            LOGGER.debug("Adding the column to the data set.")
            appui.addColumnToDataSet(alert_header_name, alert_column_index, column_bytes)
        else:
            LOGGER.debug(f"Column {alert_header_name} already exists.")
            entry = {"value": alert}
            appui.updateDataEntry("", selected_row, alert_column_index, entry)
            LOGGER.debug(f"Updated entry for row {selected_row}")
        if not StarDropHandler.column_exists(data_set, links_header_name):
            LOGGER.debug(f"Creating column {links_header_name}.")
            data_set = FastScores.construct_text_column(data_set, links_header_name)
            url = f"<a href=\"https://app.postera.ai/manifold?smiles={smiles}\">Manifold Link</a>"
            entry = app.Dataset.string(url)
            data_set.set(links_header_index, selected_row, entry)

            column_bytes = data_set.packColumn(links_header_index)
            LOGGER.debug("Adding the fast score links column to the data set.")
            appui.addColumnToDataSet(links_header_name, links_header_index, column_bytes)
        else:
            LOGGER.debug(f"Column {links_header_name} already exists")
            url = f"<a href=\"https://app.postera.ai/manifold?smiles={smiles}\">Manifold Link</a>"
            new_entry = {"value": url}
            appui.updateDataEntry("", selected_row, links_header_index, new_entry)
            LOGGER.debug(f"Updated entry for row {selected_row}")

    @staticmethod
    def add_batch_to_data_set(appui, batch_data: list, index_list: list) -> None:
        """Add the fast score data, from the batch call, to the data set.
        @param: appui: The StarDrop application interface.
        @param: batch_data: The batch to add to the data set.
        @param: index_list: The list of indexes, linking the batch data to the appropriate rows in the data set.
        """
        LOGGER.debug("Adding batch data to the data set")
        data_set_bytes = appui.importDS("")
        data_set = app.Dataset()
        data_set.parse(data_set_bytes)

        FastScores.add_batch_fast_scores_to_data_set(appui, data_set, index_list, batch_data)
        FastScores.add_batch_alert_to_data_set(appui, data_set, index_list, batch_data)
        FastScores.add_batch_links_to_data_set(appui, data_set, index_list, batch_data)

    @staticmethod
    def add_batch_fast_scores_to_data_set(appui, data_set: app.Dataset, index_list: list, batch_data: list) -> None:
        """Add the fast score value to the data set.
        @param: appui: The StarDrop application interface.
        @param: data_set: The data set to add the data too.
        @param: index_list: The list of indexes, linking the batch data to the appropriate rows in the data set.
        @param: batch_data: The batch of data retrieved from Manifold.
        """
        fast_score_header_name = "Fast SA Score"
        fast_score_column_index = data_set.columnIndex(fast_score_header_name)
        LOGGER.debug(f"The column index for {fast_score_header_name} is {fast_score_column_index}.")

        if not StarDropHandler.column_exists(data_set, fast_score_header_name):
            LOGGER.debug(f"Creating column {fast_score_header_name}")
            data_set = FastScores.construct_column(data_set, fast_score_header_name)
            for x, index in enumerate(index_list):
                entry = app.Dataset.cont(batch_data[x][1], 0.0)
                data_set.set(fast_score_column_index, index, entry)

            column_bytes = data_set.packColumn(fast_score_column_index)
            LOGGER.debug("Adding the fast score column to the data set.")
            appui.addColumnToDataSet(fast_score_header_name, fast_score_column_index, column_bytes)

        else:
            LOGGER.debug(f"Column {fast_score_header_name} already exists")
            for x, index in enumerate(index_list):
                new_entry = {"value": batch_data[x][1]}
                appui.updateDataEntry("", index, fast_score_column_index, new_entry)
            LOGGER.debug(f"Updated entries for {len(index_list)} rows.")

    @staticmethod
    def add_batch_alert_to_data_set(appui, data_set: app.Dataset, index_list: list, batch_data: list) -> None:
        """Add the fast score alert level to the data set.
        @param: appui: The StarDrop application interface.
        @param: data_set: The data set to add the data too.
        @param: index_list: The list of indexes, linking the batch data to the appropriate rows in the data set.
        @param: batch_data: The batch of data retrieved from Manifold.
        """
        alert_header_name = "Fast SA Alert"
        alert_column_index = data_set.columnIndex(alert_header_name)
        LOGGER.debug(f"The column index for {alert_header_name} is {alert_column_index}.")

        if not StarDropHandler.column_exists(data_set, alert_header_name):
            LOGGER.debug(f"Creating column {alert_header_name}")
            data_set, categories_dict = FastScores.construct_category_column(data_set, alert_header_name)
            for x, index in enumerate(index_list):
                entry = app.Dataset.createCategoryEntry(batch_data[x][2], categories_dict)
                data_set.set(alert_column_index, index, entry)

            column_bytes = data_set.packColumn(alert_column_index)
            LOGGER.debug("Adding the fast score alert column to the data set.")
            appui.addColumnToDataSet(alert_header_name, alert_column_index, column_bytes)

        else:
            LOGGER.debug(f"Column {alert_header_name} already exists")
            for x, index in enumerate(index_list):
                new_entry = {"value": batch_data[x][2]}
                appui.updateDataEntry("", index, alert_column_index, new_entry)
            LOGGER.debug(f"Updated entries for {len(index_list)} rows.")

    @staticmethod
    def add_batch_links_to_data_set(appui, data_set: app.Dataset, index_list: list, batch_data: list) -> None:
        """Add the url links retrieved from a fast score batch call, to the data set.
        @param: appui: The StarDrop application interface.
        @param: data_set: The data set to add the data too.
        @param: index_list: The list of indexes, linking the batch data to the appropriate rows in the data set.
        @param: batch_data: The batch of data retrieved from Manifold.
        """
        links_header_name = "Manifold Link"
        links_header_index = data_set.columnIndex(links_header_name)
        LOGGER.debug(f"The column index for {links_header_name} is {links_header_index}.")

        if not StarDropHandler.column_exists(data_set, links_header_name):
            LOGGER.debug(f"Creating column")
            data_set = FastScores.construct_text_column(data_set, links_header_name)
            for x, index in enumerate(index_list):
                smiles = batch_data[x][0]
                url = f"<a href=\"https://app.postera.ai/manifold?smiles={smiles}\">Manifold Link</a>"
                entry = app.Dataset.string(url)
                data_set.set(links_header_index, index, entry)

            column_bytes = data_set.packColumn(links_header_index)
            LOGGER.debug("Adding the fast score links column to the data set.")
            appui.addColumnToDataSet(links_header_name, links_header_index, column_bytes)

        else:
            LOGGER.debug(f"Column {links_header_name} already exists")
            for x, index in enumerate(index_list):
                smiles = batch_data[x][0]
                url = f"<a href=\"https://app.postera.ai/manifold?smiles={smiles}\">Manifold Link</a>"
                new_entry = {"value": url}
                appui.updateDataEntry("", index, links_header_index, new_entry)
            LOGGER.debug(f"Updated entries for {len(index_list)} rows.")

    @staticmethod
    def construct_column(data_set: app.Dataset, header_name: str, column_type = app.cont) -> app.Dataset:
        """Construct a column to hold the fast score data.
        @param: data_set: The data set to add a new column to.
        @param: header_name: The header of the new column.
        @return: The data_set with the new column added to it.
        """
        LOGGER.debug("Construct Column for fast score data.")
        row_val = data_set.rowSz()
        col_val = data_set.columnSz()
        entries = []
        header = ""

        if column_type == app.cont:
            header = data_set.createNumberHeader(header_name)
            entries.extend(app.Dataset.createInvalidEntry(app.cont) for _ in range(row_val))
        elif column_type == app.string:
            header = data_set.meta(header_name)
            entries.extend(app.Dataset.createInvalidEntry(app.string) for _ in range(row_val))
        LOGGER.debug(f"Created invalid entries for {row_val} rows.")
        data_set.insertColumn(header, entries, col_val)
        return data_set

    @staticmethod
    def construct_text_column(data_set: app.Dataset, header_name: str) -> app.Dataset:
        """Construct a column to hold the fast score links data.
        @param: data_set: The data set to add a new column to.
        @param: header_name: The header of the new column.
        @return: The data_set with the new column added to it.
        """
        LOGGER.debug("Construct Column for fast score alerts.")
        header = app.Dataset.meta(header_name)
        row_val = data_set.rowSz()
        col_val = data_set.columnSz()
        entries = []

        entries.extend(app.Dataset.createInvalidEntry(app.string) for _ in range(row_val))
        LOGGER.debug(f"Created invalid entries for {row_val} rows.")
        data_set.insertColumn(header, entries, col_val)
        return data_set

    @staticmethod
    def construct_category_column(data_set: app.Dataset, header_name: str) -> [app.Dataset, dict]:
        LOGGER.debug("Construct category column for fast score alerts.")
        categories = ['none', 'low', 'medium', 'high']
        header, categories_dict = app.Dataset.createCategoryHeader(header_name, categories)
        row_val = data_set.rowSz()
        col_val = data_set.columnSz()
        entries = []

        entries.extend(app.Dataset.createInvalidEntry(app.cat) for _ in range(row_val))
        LOGGER.debug(f"Created invalid entries for {row_val} rows.")
        data_set.insertColumn(header, entries, col_val)
        return data_set, categories_dict


class Retrosynthesis:
    @staticmethod
    def retrieve_retrosynthesis_parameters() -> dict:
        """Launch a UI element so that the user can select the parameters for the retrosynthesis API call, and return
        the values.
        @return: The parameters selected, in the form of a json.
        """
        # Create the UI element and retrieve the data from it.
        parameters = PythonQt.ManifoldRetrosynthesisParameters()
        if parameters.run():
            max_lead_time = parameters.maxLeadTime()
            reaction_tag = parameters.reactionTag()
            custom_building_blocks = parameters.customBuildingBlock()
            molecule_set_id = parameters.moleculeSetID()
            catalogs = parameters.getCatalogs()
        else:
            return {}

        # Process the data retrieved from the UI
        json_parameters = {}
        if catalogs:
            json_parameters['catalogs'] = catalogs
        if max_lead_time:
            json_parameters['maxLeadTimeWeeks'] = max_lead_time
        if reaction_tag:
            json_parameters['reactionTag'] = reaction_tag
        if custom_building_blocks:
            string_list = custom_building_blocks.split(",")
            json_parameters['customBuildingBlocks'] = string_list
        if molecule_set_id:
            json_parameters['moleculeSetIDForCustomBuildingBlocks'] = molecule_set_id
        return json_parameters

    @staticmethod
    def json_includes_error(response_json: dict) -> bool:
        """Check to see if there is an error present in the response.
        @param: response_json: The json data to check for an error.
        @return: True if there is an error, False if not.
        """
        if 'error' in response_json:
            api_error = response_json["error"]
            LOGGER.error(f"Retrosynthesis API call has returned the error: {api_error}")
            Helper.warning_message("Error!", "And error has occurred. Manifold is unable to create a molecule from the "
                                             "selected SMILES")
            return True
        else:
            LOGGER.debug("No errors were detected in the response json.")
            return False

    @staticmethod
    def process_data(response_json: dict, progress_bar) -> tuple[list, list, list]:
        """Processes the data retrieved from the Manifold API call.
        @param: response_json: The json data to process.
        @param: progress_bar: The progress bar to update as the data is processed.
        @return: A list of scores, and a list of links, from Manifold.
        """
        LOGGER.debug("Processing the data returned from Manifold.")
        if not Helper.is_api_call_valid(response_json):
            return [], [], []
        processed_score = []
        processed_min_num_steps = []
        processed_links = []
        error_messages = []

        number_of_results = len(response_json['results'])
        progress_index = 25 / number_of_results
        current_progress = 50

        for json_data in response_json['results']:
            current_progress += progress_index
            progress_bar.setProgress(current_progress)
            if 'error' in json_data:
                error_messages.append(json_data['error'])
                processed_score.append(None)
                processed_min_num_steps.append(None)
                processed_links.append(None)
            else:
                sa_data = json_data['SAData']
                processed_score.append(sa_data['score'])
                processed_min_num_steps.append(sa_data['minNumSteps'])
                processed_links.append(sa_data['manifoldLink'])

        if error_messages:
            Retrosynthesis.process_errors(error_messages)
        else:
            LOGGER.debug("No errors were detected in the response json.")
        return processed_score, processed_min_num_steps, processed_links

    @staticmethod
    def add_to_data_set(appui, scores_list: list, steps_list:list, links_list: list, index_list: list, progress_bar) -> None:
        """Add the retrosynthesis data to the data set.
        @param: appui: The StarDrop application interface.
        @param: scores_list: The list of scores to add to the data set.
        @param: links_list: The list of links to add to the data set.
        @param: index_list: The list on indexes, linking the batch data to the appropriate rows in the data set.
        """
        LOGGER.debug("Adding data to the current data set")
        data_set_bytes = appui.importDS("")
        data_set = app.Dataset()
        data_set.parse(data_set_bytes)

        progress_bar.setProgress(70)
        Retrosynthesis.add_scores_to_data_set(appui, data_set, index_list, scores_list)
        progress_bar.setProgress(80)
        Retrosynthesis.add_min_num_steps_to_data_set(appui, data_set, index_list, steps_list)
        progress_bar.setProgress(90)
        Retrosynthesis.add_links_to_data_set(appui, data_set, index_list, links_list)

    @staticmethod
    def add_scores_to_data_set(appui, data_set, index_list, scores_list) -> None:
        header_name = "Retrosynthesis SA Score"
        column_index = data_set.columnIndex(header_name)
        LOGGER.debug(f"The column index for {header_name} is {column_index}.")

        if not StarDropHandler.column_exists(data_set, header_name):
            LOGGER.debug(f"Creating column for {header_name}.")
            data_set = Retrosynthesis.construct_column(data_set, header_name, app.cont)
            for x, index in enumerate(index_list):
                if scores_list[x] is not None:
                    entry = app.Dataset.cont(scores_list[x], 0.0)
                    data_set.set(column_index, index, entry)
            column_bytes = data_set.packColumn(column_index)
            LOGGER.debug(f"Adding the column {header_name} to the data set.")
            appui.addColumnToDataSet(header_name, column_index, column_bytes)
        else:
            LOGGER.debug(f"Column {header_name} already exists.")
            for x, index in enumerate(index_list):
                if scores_list[x] is not None:
                    new_entry = {"value": scores_list[x]}
                    appui.updateDataEntry("", index, column_index, new_entry)

    @staticmethod
    def add_links_to_data_set(appui, data_set, index_list, links_list) -> None:
        header_name = "Manifold Link"
        links_column_index = data_set.columnIndex(header_name)
        LOGGER.debug(f"The column index for {header_name} is {links_column_index}.")

        if not StarDropHandler.column_exists(data_set, header_name):
            LOGGER.debug(f"Creating column for {header_name}.")
            data_set = Retrosynthesis.construct_column(data_set, header_name, app.string)
            for x, index in enumerate(index_list):
                if links_list[x] is not None:
                    url = f"<a href=\"{links_list[x]}\">Manifold Link</a>"
                    LOGGER.debug(f"The url for row {index} is {links_list[x]}.")
                    entry = app.Dataset.string(url)
                    data_set.set(links_column_index, index, entry)

            column_bytes = data_set.packColumn(links_column_index)
            LOGGER.debug(f"Adding the column '{header_name}' to the data set.")
            appui.addColumnToDataSet(header_name, links_column_index, column_bytes)
        else:
            LOGGER.debug(f"Column {header_name} already exists.")
            for x, index in enumerate(index_list):
                if links_list[x] is not None:
                    url = f"<a href=\"{links_list[x]}\">Manifold Link</a>"
                    LOGGER.debug(f"The url for row {index} is {links_list[x]}.")
                    new_entry = {"value": url}
                    LOGGER.debug(f"The url for row {index} is {links_list[x]}.")
                    appui.updateDataEntry("", index, links_column_index, new_entry)

    @staticmethod
    def add_min_num_steps_to_data_set(appui, data_set, index_list, steps_list) -> None:
        header_name = "Retrosynthesis Min Num Steps"
        steps_column_index = data_set.columnIndex(header_name)
        LOGGER.debug(f"The column index for {header_name} is {steps_column_index}")

        if not StarDropHandler.column_exists(data_set, header_name):
            LOGGER.debug(f"Creating column for {header_name}")
            data_set = Retrosynthesis.construct_column(data_set, header_name, app.cont)
            for x, index in enumerate(index_list):
                if steps_list[x] is not None:
                    entry = app.Dataset.cont(steps_list[x], 0.0)
                    data_set.set(steps_column_index, index, entry)
            column_bytes = data_set.packColumn(steps_column_index)
            LOGGER.debug(f"Adding the column {header_name} to the data_set.")
            appui.addColumnToDataSet(header_name, steps_column_index, column_bytes)
        else:
            LOGGER.debug(f"Column {header_name} already exists.")
            for x, index in enumerate(index_list):
                if steps_list[x] is not None:
                    new_entry = {"value": steps_list[x]}
                    appui.updateDataEntry("", index, steps_column_index, new_entry)

    @staticmethod
    def construct_column(data_set: app.Dataset, header_name: str, column_type):
        """Construct a column to hold the retrosynthesis data.
        @param: data_set: The data set to add a new column to.
        @param: header_name: The header of the new column.
        @param: column_type: The type of column that needs to be made.
        @return: The data_set with the new column added to it.
        """
        LOGGER.debug("Construct Column")
        row_val = data_set.rowSz()
        col_val = data_set.columnSz()
        entries = []

        if column_type == app.string:
            header = app.Dataset.meta(header_name)
            entries.extend(app.Dataset.createInvalidEntry(app.string) for _ in range(row_val))
        elif column_type == app.cont:
            header = data_set.createNumberHeader(header_name)
            entries.extend(app.Dataset.createInvalidEntry(app.cont) for _ in range(row_val))
        else:
            return None
        LOGGER.debug(f"Created invalid entries for {row_val} rows.")
        data_set.insertColumn(header, entries, col_val)
        return data_set

    @staticmethod
    def process_errors(error_messages: list) -> None:
        """Process the errors that have occurred during the retrosynthesis API call.
        @param: error_messages: The list of errors to process.
        """
        error_count = 0
        for msg in error_messages:
            LOGGER.error(msg)
            if "Retrosynthesis time out for" in msg:
                error_message = "Retrosynthesis API call timed out. Please retry your query."
                Helper.warning_message(title="API Timeout",
                                       message=error_message)
                return None
            else:
                error_count += 1

        if error_count == 1:
            warning_message = f"{error_count} entry has an invalid SMILES string."
        else:
            warning_message = f"{error_count} entries have invalid SMILES strings."

        Helper.warning_message(message=warning_message)


class Helper:
    @staticmethod
    def warning_message(title: Optional[str] = "Warning",
                        message: Optional[str] = "Something unexpected has occurred.") -> None:
        """Generate a warning message, and display it in a dialog.
        @param: title: The title of the dialog.
        @param: message: The body of the warning message.
        @return: True is the user pressed the OK button, False if not.
        """
        LOGGER.warning(message)
        message_dialog = PythonQt.Warning()
        message_dialog.setWindowTitle(title)
        message_dialog.setMessage(message)
        return message_dialog.run()

    @staticmethod
    def connect_to_manifold(appui) -> bool:
        """Check that the user has a token set.
        @param: appui: The StarDrop application interface.
        @returns: True if a token has been set, False otherwise.
        """
        token = manifoldconfig.get_token()
        LOGGER.info('Connecting to vault')
        if token == '':
            LOGGER.debug("No token set, changing token.")
            change_token(appui)
            if token == '':
                LOGGER.debug("Unable to connect to vault due to missing token.")
                return False
        return True

    @staticmethod
    def extract_selected_smiles_string(appui) -> Union[tuple[str, int], tuple[None, None]]:
        """Extract the SMILES string of the selected row.
        @param: appui: The StarDrop application interface.
        @return: A single SMILES string and the row it was extracted from
        """
        selected_molecule = appui.getCurrentMolecule()
        # We want the first selected row,
        #  as we are only getting data from one row.
        selected_rows = appui.getSelectedRows("")
        if selected_rows:
            selected_row = selected_rows[0]
            LOGGER.debug(f"Extracted the following SMILES string from the data set: {selected_molecule}, "
                         f"from row {selected_row}")
            return selected_molecule, selected_row
        return None, None

    @staticmethod
    def extract_selected_smiles_strings(appui) -> tuple[list, list]:
        """Extract the selected smiles from the data set.
        @param: appui: The StarDrop application interface.
        @return: A list of SMILES strings, and a list of row indexes
        """
        data_set = app.Dataset()
        data_set.parse(appui.importDS(""))
        selected_rows = appui.getSelectedRows("")
        num_of_cols = data_set.columnSz()

        structure_column_index = -1
        for i in range(num_of_cols):
            if data_set.columnType(i) == app.mol:
                structure_column_index = i
                break
        LOGGER.debug(f"The column index of the structures is: {structure_column_index}")

        list_of_smiles = []
        list_of_indexes = []
        for row in selected_rows:
            smiles_string = data_set.stringval(structure_column_index, row)
            list_of_smiles.append(smiles_string)
            list_of_indexes.append(row)

        LOGGER.debug(f"Extracted {len(list_of_smiles)} smiles and {len(list_of_indexes)} indexes.")

        return list_of_smiles, list_of_indexes

    @staticmethod
    def perform_post_request(api_url: str, json_data: dict) -> dict:
        """Perform a post request, using the parameters passed.
        @param: API_url: The url to send the request to.
        @param: json_data: The data to pass as part of the request.
        @return: The response data, as a json document.
        """
        LOGGER.debug(f"Performing Post Request to: {api_url}")
        try:
            response = requests.post(
                url=api_url,
                headers={
                    'X-API-KEY': manifoldconfig.get_token(),
                },
                json=json_data
            )
        except requests.ConnectionError:
            LOGGER.error("Connection Error: Unable to connect to the Manifold API")
            Helper.warning_message("Error!", "An error has occurred. StarDrop is unable to connect to Manifold.")
            return {}
        except Exception as e:
            LOGGER.error("Error Message: {msg}".format(msg=e))
            Helper.warning_message("Error!", "An error has occurred. "
                                             "StarDrop is unable to retrieve data from Manifold.")
            return {}
        return response.json()

    @staticmethod
    def distribute_list_into_chunks(old_list: list, chunk_size: int) -> list:
        """Take a list and split it into multiple lists, each a maximum of 'chunk_size'.
        @param: old_list: The list we want to split.
        @param: chunk_size: The maximum size of each new list.
        @yield: A new list, containing a maximum of 'chunk_size' elements.
        """
        for i in range(0, len(old_list), chunk_size):
            yield old_list[i:i + chunk_size]

    @staticmethod
    def check_for_exceeded_api_usage(json_data: dict) -> bool:
        """Checks the json response dictionary, to see if the message for exceeding the API usage is being received.
        @param: json_data: The dictionary to check for the API message.
        @return: True if the json includes the message, False if not.
        """
        if 'detail' in json_data:
            LOGGER.critical(f"The user has reached their maximum daily API queries. "
                            f"API message is: {json_data['detail']}")
            Helper.warning_message(title="Daily API Queries",
                                   message="You have reached your daily amount of API queries."
                                           "\nPlease try again tomorrow.")
            return True
        return False

    @staticmethod
    def is_api_call_valid(json_data: dict) -> bool:
        """Checks the json response, to see if there are any issues with the API call.
        @param: json_data: The json to check for the API issue message.
        @return: True if the API call was valid, and returned no initial issue.

        NOTE: This does not check for errors, just a detail message returned from Manifold
        """
        if 'detail' in json_data:
            if json_data['detail'] == "Invalid api key provided":
                LOGGER.critical(f"The user is using an invalid API key.")
                Helper.warning_message(title="Invalid API key",
                                       message="The API key that you have entered is invalid.\nEnsure that you have "
                                               "input a valid API key.")
                return False
            if all(x in json_data['detail'] for x in ["Already made", "queries this day"]):
                LOGGER.critical(f"The user has reached their maximum daily API queries. "
                                f"API message is: {json_data['detail']}")
                Helper.warning_message(title="Daily API Queries",
                                       message="You have reached your daily amount of API queries."
                                               "\nPlease try again tomorrow or contact PostEra.")
                return False
        return True


class StarDropHandler:
    @staticmethod
    def column_exists(data_set, column_name: str) -> bool:
        """Check to see if the column already exists in the data set.
        @param: data_set: The data set to check for the column in.
        @param: column_name: The column header to look for.
        @return: True if the column exists, False if not.
        """
        LOGGER.debug("Check if column exists")
        col_index = data_set.columnIndex(column_name)
        if col_index == -1:
            LOGGER.debug("Column doesn't exist.")
            return False
        LOGGER.debug("Column does exist.")
        return True
